#! /usr/bin/env python
# -*- coding: utf-8 -*-

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#   see <https://www.gnu.org/licenses/>.

#  GIMP plugin to create tartan patterns
# (c) Ofnuts 2021
#
#   History:
#
#   v0.0: 2021-06-07: First published version
#   v0.1: 2021-06-08: Handle non-RGB images
#                     Auto-fix aspect ratio
#                     Weave direction
#                     Scale
#   v0.2: 2021-06-09: Improve speed by painting the full mask directly
#                     Fix image names
#   v0.3: 2021-06-09: Even faster mask painting
#                     `None` weaving option

# Scottish fabric: https://www.gimp-forum.net/Thread-Scottish-fabric?pid=23836#pid23836
# Tartan (a.k.a. plaid) cloth in a few easy steps (tutorial): https://www.gimp-forum.net/Thread-Tartan-a-k-a-plaid-cloth-in-a-few-easy-steps?pid=24000#pid24000
# ofn-tartan: https://www.gimp-forum.net/Thread-ofn-tartan
# Download source: https://sourceforge.net/projects/gimp-tools/files/scripts/

# (c) MareroQ     
#   v0.3 with preview: 2021-07-12: add "Preview" by fu_preview 0.1.1 (modificed gimpfu) by Claude Lion [aka cli345] http://gimpchat.com/viewtopic.php?f=9&t=19420
#   v0.4 with preview: add tabs to make more clear the interface: PF_TAB and PF_TITLE based by code J.F.Garcia [aka: arakne] gimpfu3.py (in follow-path5) https://www.arakne.es/en/gimp-plugin-follow-path-map-values-on-the-y-coordinate/
#   v0.5 Previous versions it only allowed to create a certain initial size width=100 px (square seamless pattern=200 px).
#        The limit to 256 px has been changed (and the end width of the square seamless pattern is 2*the maximum size of the "End" item).
#   v0.6 fix the error for odd values: limit the acceptable values of START - END to even values  

#####################################################################################################################################################
############################################# START fu_preview 0.1.1 by Claude Lion $ J.F. Garcia ###################################################
#####################################################################################################################################################

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb
import os, time, sys
import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}


def register(proc_name, blurb, help, author, copyright, date, label,
             imagetypes, params, results, function,
             menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
                return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements "
                          "(%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead"
                       % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright,
                                       date, label, imagetypes,
                                       plugin_type, params, results,
                                       function, menu, domain,
                                       on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date,
         label, imagetypes, plugin_type,
         params, results, function, menu, domain,
         on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]],
                     x[1],
                     _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode",
                                     "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright,
                               date, label, imagetypes, plugin_type,
                               params, results)

        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date,
     label, imagetypes, plugin_type,
     params, results, function, menu, domain,
     on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
#    import pango
    gimpui.gimp_ui_init () # add MrQ
    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default="", **kwargs): # modif C.L.
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            self.set_activates_default(True)
            #~ self.connect('paste-clipboard', lambda _: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            self.get_buffer().connect('inserted-text', lambda _1, _2, _3, _4: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            self.get_buffer().connect('deleted-text', lambda _1, _2, _3: kwargs['dialg'].la_previ.invalidate()) # modif C.L

        def get_value(self):
            return self.get_text()


    class LabelTitle(gtk.Label):
        #def __init__ (self, default="", **kwargs): # modif C.L. ???
        def __init__(self, default="", align = (0.0, 0.0)):		
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default="", **kwargs): # modif C.L.
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))
            self.buffer.connect('changed', lambda _: kwargs['dialg'].la_previ.invalidate()) # modif C.L.			

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5), **kwargs): # modif C.L.
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1],
                                      step, 10 * step, 0)
            # modif C.L.
            def react(adj):
                return self.get_toplevel().la_previ.invalidate()
            #~ self.adj.connect('value-changed', lambda adj: self.get_toplevel().la_previ.invalidate())
            self.adj.connect('value-changed', react)
            # end modif C.L.
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=10, bounds=(0, 100, 5), **kwargs): # modif C.L.
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1],
                                      step, 10 * step, 0)
            self.adj.connect('value-changed', lambda adj: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))


    class ToggleEntry(gtk.ToggleButton):
        #~ def __init__(self, default=0):
        def __init__(self, default=0,labels=[_("No"),_("Yes")], **kwargs): # modif C.L.
            
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()

            #~ self.connect("toggled", self.changed)
            # modif C.L.
            def react(tog):
                self.changed(tog)
                return kwargs['dialg'].la_previ.invalidate()
            self.connect("toggled", react)
            # end modif C.L.


            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()



    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0)), **kwargs): # modif C.L.
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=(), **kwargs): # modif C.L.
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)
            self.connect('changed', lambda adj: kwargs['dialg'].la_previ.invalidate()) # modif C.L.

        def get_value(self):
            return self.get_active()

    def FileSelector(default="", title=None, **kwargs): # modif C.L.
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            if default == "/": default = ""
            return DirnameSelector(default)
        else:
            return FilenameSelector(default, title=title, save_mode=False)

    class FilenameSelector(gtk.HBox):
        #gimpfu.FileChooserButton
        def __init__(self, default, save_mode=True, title=None, **kwargs): # modif C.L.
            super(FilenameSelector, self).__init__()
            if not title:
                self.title = _("Python-Fu File Selection")
            else:
                self.title = title
            self.save_mode = save_mode
            box = self
            self.entry = gtk.Entry()
            image = gtk.Image()
            image.set_from_stock(gtk.STOCK_FILE, gtk.ICON_SIZE_BUTTON)
            self.button = gtk.Button()
            self.button.set_image(image)
            box.pack_start(self.entry)
            box.pack_start(self.button, expand=False)
            self.button.connect("clicked", self.pick_file)
            if default:
                self.entry.set_text(default)

        def show(self):
            super(FilenameSelector, self).show()
            self.button.show()
            self.entry.show()

        def pick_file(self, widget):
            entry = self.entry
            dialog = gtk.FileChooserDialog(
                         title=self.title,
                         action=(gtk.FILE_CHOOSER_ACTION_SAVE
                                     if self.save_mode else
                                 gtk.FILE_CHOOSER_ACTION_OPEN),
                         buttons=(gtk.STOCK_CANCEL,
                                gtk.RESPONSE_CANCEL,
                                gtk.STOCK_SAVE
                                     if self.save_mode else
                                gtk.STOCK_OPEN,
                                gtk.RESPONSE_OK)
                        )
            dialog.set_alternative_button_order ((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))
            dialog.show_all()
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                entry.set_text(dialog.get_filename())
            dialog.destroy()

        def get_value(self):
            return self.entry.get_text()


    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default="", **kwargs): # modif C.L.
            gtk.FileChooserButton.__init__(self,
                                           _("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    # modif C.L.
    class ColorSelector(gimpui.ColorSelector):
        def __init__(self, default=gimpcolor.RGB(1.0, 0, 0), **kwargs):
            ret = gimpui.ColorSelector.__init__(self, default)
            self.connect('color-changed', lambda _: kwargs['dialg'].la_previ.invalidate())
            return ret
    class FontSelector(gimpui.FontSelector):
        def __init__(self, default="Sans", **kwargs):
            ret = gimpui.FontSelector.__init__(self, default)
            self.connect('font-set', lambda _1, _2, _3: kwargs['dialg'].la_previ.invalidate())
            return ret
    class PatternSelector(gimpui.PatternSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.PatternSelector.__init__(self, default)
            self.connect('pattern-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    class GradientSelector(gimpui.GradientSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.GradientSelector.__init__(self, default)
            self.connect('gradient-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    class PaletteSelector(gimpui.PaletteSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.PaletteSelector.__init__(self, default)
            self.connect('palette-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    # end modif C.L.
    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : ColorSelector, # modif C.L.
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : PatternSelector, # modif C.L.
            PF_GRADIENT    : GradientSelector, # modif C.L.
            PF_PALETTE     : PaletteSelector, # modif C.L.
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    # modif C.L.
    hboxPrincipal = gtk.HBox(spacing=10)
	
    dialog.vbox.pack_start(hboxPrincipal)			        


    hboxPrincipal.show()
    
    dialog.la_previ = gimpui.DrawablePreview(start_params[1])
    dialog.la_previ.set_size_request(400,400)               # New Line MrQ	
    dialog.la_previ.la_img = start_params[0]
    dialog.la_previ.la_layer = start_params[1]
    hboxPrincipal.pack_start(dialog.la_previ)

    dialog.set_default_size(1100, 600)  					# New Line MrQ
    dialog.la_previ.show()
    # end modif C.L.



    vbox = gtk.VBox(False, 2)
    vbox.set_border_width(2)
    vbox.set_size_request(480, 0)                           # modif C.L. + MrQ
    #~ dialog.vbox.pack_start(vbox)
    hboxPrincipal.pack_start(vbox) # modif C.L.
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})
    # modif C.L.
    def ifinvalidated(p_la_previ):
        
        # get dimension
        #===============
        x, y = p_la_previ.get_position()
        w, h = p_la_previ.get_size()
        
        # Effet
        #=========
        # has alpha
        lb_alpha = pdb.gimp_drawable_has_alpha(p_la_previ.la_layer)
        
        # img, layer temporaires
        tempImg = pdb.gimp_image_new(w, h, 0)
        #~ tempLayer = pdb.gimp_layer_new(tempImg, w, h, 0, 'temp', 100, 0)
        tempLayer = pdb.gimp_layer_new(tempImg, w, h, lb_alpha, 'temp', 100, 0)
        pdb.gimp_image_add_layer(tempImg, tempLayer, 0)
        
        # copie orig dans tempo
        rgnPreviIn = p_la_previ.la_layer.get_pixel_rgn(x, y, w, h)
        rgnTemp = tempLayer.get_pixel_rgn(0, 0, w, h)
        rgnTemp[0:w, 0:h] = rgnPreviIn[x:x+w, y:y+h]
        
        # effect
        tempImg.undo_group_start()
        tempImg.disable_undo()
        
        params = []

        try:
            for wid in edit_wids:
                params.append(wid.get_value())
        except EntryValueError:
            warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
        else:
            try:
                paramsTot = (tempImg, tempLayer) + tuple(params)
                _set_defaults(proc_name, start_params + tuple(params))
                function(*paramsTot)
                
                res = tempImg.active_layer
                
                if not lb_alpha: pdb.gimp_layer_flatten(res)
                
                tempImg.undo_group_end()
                
                rgnRes = res.get_pixel_rgn(0, 0, w, h)
                
                # mise dans previ
                #================
                p_la_previ.draw_buffer(rgnRes[0:w, 0:h], int(3+lb_alpha)*w)
                
                pdb.gimp_image_delete(tempImg)
                
                #pdb.gimp_display_new(tempImg)
            except CancelError:
                pass
            except Exception as e:
                dialog.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                error_dialog(dialog, proc_name)
                raise
    # end modif C.L.
    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()
    dialog.la_previ.connect('invalidated', ifinvalidated) # modif C.L.
    dialog.connect("response", response)
    iniX=0
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]

        tab, incr, rows, connect, cols = (0, 1, 1, None, 1)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE:
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW):
                wid = _edit_mapping[pf_type](def_val, params[i][4], dialg=dialog) # modif C.L.
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4],dialg=dialog)
                else:
                    wid = _edit_mapping[pf_type](def_val)				


            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val, dialg=dialog) # modif C.L.
            if pf_type!=PF_TITLE:
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', desc)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()
	

	
    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break

        min_args = i

    if len(script_params) > min_args:
        start_params = params[:min_args + 1]

        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE

    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])

    gimp.displays_flush()

    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message
	
	
#####################################################################################################################################################
############################################# END fu_preview 0.1.1 by Claude Lion $ J.F. Garcia #####################################################
#####################################################################################################################################################


version='0.3 with preview'
import math
import sys,os,time
from collections import namedtuple


  
def trace(format,*args):
    if debug:
        print format % args

def createOpts(name,pairs):
    optsclass=namedtuple(name+'Type',[symbol for symbol,label in pairs]+['labels','labelTuples'])
    opts=optsclass(*(
                    range(len(pairs))
                    +[[label for symbol,label in pairs]]
                    +[[(label,i) for i,(symbol,label) in enumerate(pairs)]]
                    ))
    return opts

Second=createOpts('Second',[('REVERSE','Reverse'),('REPEAT','Repeat')])
Weave=createOpts('Weave',[('NONE','None, transparency'),('NWSE','NW➤SE'),('NESW','NE➤SW')])


def removeAllGuides(image):
    g=pdb.gimp_image_find_next_guide(image,0)
    while g:
        pdb.gimp_image_delete_guide(image,g)
        g=pdb.gimp_image_find_next_guide(image,0)

def squareLayer(tileImage,layer,reverse,name):
    firstPart=layer.copy()
    firstPart.name='# %s #' % name
    tileImage.add_layer(firstPart,0)
    secondPart=layer.copy()
    tileImage.add_layer(secondPart,0)
    if reverse==Second.REVERSE:
        secondPart.transform_flip_simple(ORIENTATION_HORIZONTAL,True,0)
    secondPart.set_offsets(layer.width,0)
    return tileImage.merge_down(secondPart,CLIP_TO_IMAGE)

def addMask(layer,weave,shift):
    base='\x00\x00\xff\xff'*(2+layer.width/4)
    rows=[]
    for i in range(4):
        rows.append(base[i+shift:i+shift+layer.width])
    if weave==Weave.NWSE:
        rows.reverse();
    maskBytes=(''.join(rows))*(layer.height/4)
    mask=layer.create_mask(ADD_MASK_WHITE)
    layer.add_mask(mask)
    region=mask.get_pixel_rgn(0, 0, mask.width,mask.height)
    region[:,:]=maskBytes
    
def createTileImage(imageName,layer,settImage,warp2nd,weft2nd,weave,shift,scale):
    #tileImage=settImage.duplicate() MrQ
    tileImage=pdb.gimp_item_get_image(layer)
    tileImage.filename=os.path.join(gimp.user_directory(4),"Tartan tile from %s" % imageName)
    pdb.gimp_image_undo_disable(tileImage)
    if tileImage.base_type!=RGB:
         pdb.gimp_image_convert_rgb(tileImage)
    pdb.gimp_selection_none(tileImage)
    removeAllGuides(tileImage)
    #settLayer=tileImage.flatten()
    # Fix aspect ratio
    #pdb.gimp_image_scale_full(tileImage,tileImage.width,tileImage.width*2,INTERPOLATION_NONE)
    #settLayer.name='# Sett #'
    #tileImage.resize(2*tileImage.width,tileImage.height,0,0)

    weftLayer=squareLayer(tileImage,layer,weft2nd,'weft')
    weftLayer.transform_rotate_simple(ROTATE_90,True,0,0,True)
    warpLayer=squareLayer(tileImage,layer,warp2nd,'warp')
    if weave==Weave.NONE:
        warpLayer.opacity=50.
    else:
        addMask(warpLayer,weave,shift)
    tileImage.remove_layer(layer)
    # Scale
    #pdb.gimp_image_scale_full(tileImage,tileImage.width*scale,tileImage.height*scale,INTERPOLATION_NONE)
    pdb.gimp_image_undo_enable(tileImage)
    tileImage.clean_all()
    #gimp.Display(tileImage)
    return tileImage
	
'''
def createTestImage(imageName,width,height):
    testImage=gimp.Image(width,height,RGB)
    testImage.filename=os.path.join(gimp.user_directory(4),"Tartan test from %s" % imageName)
    pdb.gimp_image_undo_disable(testImage)
    testLayer=gimp.Layer(testImage,"Tartan from %s" % imageName,width,height,RGB,100,NORMAL_MODE)
    setClipboardPattern()
    pdb.gimp_drawable_fill(testLayer,FILL_PATTERN)
    testImage.add_layer(testLayer,0)
    pdb.gimp_image_undo_enable(testImage)
    testImage.clean_all()
    gimp.Display(testImage)
    return testImage
 '''   
def ofn_tartan_preview_tab(image,layer,background,color_1,color_2,sx1,sx2,ex1,ex2,color_3,color_4,sx3,sx4,ex3,ex4,color_5,color_6,sx5,sx6,ex5,ex6, warp2nd,weft2nd,weave,shift, render, fill_layer, help, ptab):
    
    for lay in image.layers:
        pdb.gimp_image_set_active_layer(image,lay)
        pdb.gimp_item_set_visible(lay, 0)
		
    pdb.gimp_selection_none(image)
	
    if fill_layer==1:

       active_layer = pdb.gimp_layer_new_from_drawable(layer, image)
       pdb.gimp_image_insert_layer(image, active_layer, None, -1)
       active_layer.name = layer.name   
       pdb.gimp_item_set_visible(active_layer, False)
	   
    if ex1-sx1>=0:
	   x1=ex1-sx1
    else:
       x1=0	
	   
    if ex2-sx2>=0 :
	   x2=ex2-sx2
    else:
       x2=0	
	   
    if ex3-sx3>=0 :
	   x3=ex3-sx3
    else:
       x3=0	
	   
    if ex4-sx4>=0 :
	   x4=ex4-sx4
    else:
       x4=0	
	   
    if ex5-sx5>=0 :
	   x5=ex5-sx5
    else:
       x5=0	
	   
    if ex6-sx6>=0 :
	   x6=ex6-sx6
    else:
       x6=0		   
    sx=max(ex1,ex2,ex3,ex4,ex5,ex6) 
    #sx=x1 + x2 + x3 + x4 + x5 + x6
    sy=sx*2	
    tartan = pdb.gimp_layer_new(image, sx, sy, RGBA_IMAGE, "Tartan", 100, LAYER_MODE_NORMAL_LEGACY)
    pdb.gimp_image_insert_layer(image, tartan, None, 0)
    layer.add_alpha()
    if background==0:	# BACKGROUND
        pdb.gimp_context_get_background()
        pdb.gimp_drawable_edit_fill(tartan, 1)
    if background==1:   # TRANSPARENT
        pdb.gimp_context_get_background()
        pdb.gimp_drawable_edit_fill(tartan, 3)

	
    if ex1 > 0  and ex1 > sx1:
        pdb.gimp_image_select_rectangle(image, CHANNEL_OP_REPLACE, sx1, 0, ex1-sx1, sy) 
        pdb.gimp_context_set_foreground(color_1)		
        pdb.gimp_drawable_edit_fill(tartan, FILL_FOREGROUND)
        pdb.gimp_selection_none(image)		
    if ex2 > 0  and ex2 > sx2:
        pdb.gimp_image_select_rectangle(image, CHANNEL_OP_REPLACE, sx2, 0, ex2-sx2, sy) 
        pdb.gimp_context_set_foreground(color_2)		
        pdb.gimp_drawable_edit_fill(tartan, FILL_FOREGROUND)
        pdb.gimp_selection_none(image)
    if ex3 > 0  and ex3 > sx3:
        pdb.gimp_image_select_rectangle(image, CHANNEL_OP_REPLACE, sx3, 0, ex3-sx3, sy) 
        pdb.gimp_context_set_foreground(color_3)		
        pdb.gimp_drawable_edit_fill(tartan, FILL_FOREGROUND)
        pdb.gimp_selection_none(image)	
    if ex4 > 0 and ex4 > sx4:	
        pdb.gimp_image_select_rectangle(image, CHANNEL_OP_REPLACE, sx4, 0, ex4-sx4, sy) 
        pdb.gimp_context_set_foreground(color_4)		
        pdb.gimp_drawable_edit_fill(tartan, FILL_FOREGROUND)
        pdb.gimp_selection_none(image)		
    if ex5 > 0 and ex5 > sx5:	
        pdb.gimp_image_select_rectangle(image, CHANNEL_OP_REPLACE, sx5, 0, ex5-sx5, sy) 
        pdb.gimp_context_set_foreground(color_5)		
        pdb.gimp_drawable_edit_fill(tartan, FILL_FOREGROUND)
        pdb.gimp_selection_none(image)		
    if ex6 > 0 and ex6 > sx6:	
        pdb.gimp_image_select_rectangle(image, CHANNEL_OP_REPLACE, sx6, 0, ex6-sx6, sy) 
        pdb.gimp_context_set_foreground(color_6)		
        pdb.gimp_drawable_edit_fill(tartan, FILL_FOREGROUND)		
        pdb.gimp_selection_none(image)
    scale=1
    shift,scale=map(int,(shift,scale))

    if render==1: 
	
       tileImage=createTileImage(image.name,tartan,image,warp2nd,weft2nd,weave,shift,scale)
       pdb.gimp_item_set_visible(layer, False)	   
       vis=pdb.gimp_image_merge_visible_layers(image, 0)
       vis.name="Tartan"
       pdb.gimp_edit_copy(vis)
	   
    if fill_layer==1:
	   
       pdb.gimp_drawable_edit_fill(layer,FILL_PATTERN)
       pdb.gimp_image_raise_item_to_top(image, layer)
       pdb.gimp_item_set_visible(layer, True)
	   
    view = pdb.gimp_layer_new_from_visible(image, image, "Preview from Visible") 
    pdb.gimp_image_insert_layer(image, view, None, 0)	 
    gimp.displays_flush()
    
### Registrations

# Some variables to avoid repeating myself
desc="Tartan Ver.0.6 with Preview"
whoiam='\n%s (v%s)' % (os.path.abspath(sys.argv[0]),version)
author='original code:Ofnuts, modification: MareroQ'
year='2021'
hlp = "<b> Plugin generates tartan patterns</b>.\n\n Code and description from ofn-tartan.py.\n The side length of the active layer must be greater than 512 px."\
+ "\n From the initial layer where: width= max. End (0-256 px) hight=2*width  \n it 'Render' a square seamless pattern 2*width px."\
+ "\n The Start and End positions are counted according to the X axis."\
+ "\n The script creates the warp and weft image by:"\
+ "\n - enlarging the initial layer (by 'Reverse' or 'Repeat' method),\n - duplicating this layer and forming the warp and weft\n  by rotating the warp image by 90 ° and adding a weaving mask."\
+ "\n\n Parameters:\n"\
+ "\n	*Warp, Weft*\n Indicates whether the second warp and weft portion is obtained by inverting\n or repeating dice (traditional tartan uses inversion)"\
+ "\n	*Weave*\n The direction of the weave pattern.\n If None, then the weaving pattern is replaced with 50% opacity."\
+ "\n	*Thread Offset*\n Enables the warping mask (weaving) to be offset.\n Sometimes this gives a better look to very thin stripes (one or two pixels wide)."

register('ofn-tartan-preview-tab',
         desc,
         desc,
         author,"GPL3",year,
         "Create Tartan with Preview...",
         'RGB*',
         [
             (PF_IMAGE,'image','Input image',None),
             (PF_DRAWABLE, "drawable", "Input drawable", None), # MrQ
             (PF_OPTION,"background", "Color Primer", 0,("FILL-BACKGROUND","FILL-TRANSPARENT"),{'newLine':'1','tab':0}),				 
             (PF_COLOR, "color_1", "Color 1", (16,16,16),0,{'newLine':'0','tab':0}),
             (PF_COLOR, "color_2", "	Color 2", (232,190,0),0,{'newLine':'1','tab':0}),
             (PF_SPINNER, "sx1", "	Start:", 0, (0, 256, 2),{'newLine':'0','tab':0}),		 
             (PF_SPINNER, "sx2", "		Start:", 16, (0, 256, 2),{'newLine':'1','tab':0,'iniCol':2}),
             (PF_SPINNER, "ex1", "	End:", 16, (0, 256, 2),{'newLine':'0','tab':0}),			 
             (PF_SPINNER, "ex2", "		End:", 20, (0, 256, 2),{'newLine':'1','tab':0,'iniCol':2}),	
			 
             (PF_COLOR, "color_3", "Color 3", (16,16,16),0,{'newLine':'0','tab':0}),
             (PF_COLOR, "color_4", "	Color 4", (232,190,0),0,{'newLine':'1','tab':0}),
			 
             (PF_SPINNER, "sx3", "	Start:", 20, (0, 256, 2),{'newLine':'0','tab':0}),
             (PF_SPINNER, "sx4", "		Start:", 52, (0, 256, 2),{'newLine':'1','tab':0,'iniCol':2}),			 
             (PF_SPINNER, "ex3", "	End:", 52, (0, 256, 2),{'newLine':'0','tab':0}),
             (PF_SPINNER, "ex4", "		End:", 100, (0, 256, 2),{'newLine':'1','tab':0,'iniCol':2}),
			 
             (PF_COLOR, "color_5", "Color 5", (200, 0, 0),0,{'newLine':'0','tab':0}),
             (PF_COLOR, "color_6", "	Color 6", (145, 38, 142),0,{'newLine':'1','tab':0}),
			 
             (PF_SPINNER, "sx5", "	Start:", 100, (0, 256, 2),{'newLine':'0','tab':0}),
             (PF_SPINNER, "sx6", "		Start:", 0, (0, 256, 2),{'newLine':'1','tab':0,'iniCol':2}),			 
             (PF_SPINNER, "ex5", "	End:", 104, (0, 256, 2),{'newLine':'0','tab':0}),
             (PF_SPINNER, "ex6", "		End:", 0, (0, 256, 2),{'newLine':'1','tab':0,'iniCol':2}),
			 
             (PF_OPTION,'warp2nd','Warp',Second.REVERSE,Second.labels,{'newLine':'1','tab':1}),
             (PF_OPTION,'weft2nd','Weft',Second.REVERSE,Second.labels,{'newLine':'1','tab':1}),
             (PF_OPTION,'weave','Weave',Weave.NESW,Weave.labels,{'newLine':'1','tab':1}),
             (PF_SPINNER,'shift','Thread shift',0,(0,3,1),{'newLine':'1','tab':1}),
             (PF_TOGGLE, "render",   "Render Tartan:", 0,[("Don't Use "),(" YES")],{'newLine':'1','tab':1}),
             (PF_TOGGLE, "fill_layer",   "Fill active layer:", 0,[("Don't Use "),(" YES")],{'newLine':'1','tab':1}),

             (PF_TITLE  , "help"  , hlp ,(10,10), None           ,{'tab':2}),		
             (PF_TAB    , "ptab"    , "tabs" , 0, (' Coloring ',' Tartan ',' HELP ',) ),			 
            ],         
         [],
         ofn_tartan_preview_tab,
         menu='<Image>/Filters/Render/Pattern'
         )
main()
